/*
 * Andrea Di Biagio
 * Politecnico di Milano, 2007
 * 
 * axe_gencode.h
 * Formal Languages & Compilers Machine, 2007/2008
 * 
 */

#ifndef _AXE_GENCODE_H
#define _AXE_GENCODE_H

#include "axe_engine.h"
#include "axe_struct.h"

/*----------------------------------------------------
 *                   NOP & HALT
 *---------------------------------------------------*/

/* By calling this function, a new NOP instruction will be added
 * to `program'. A NOP instruction doesn't make use of
 * any kind of parameter */
extern t_axe_instruction * gen_nop_instruction
      (t_program_infos *program);

/* By calling this function, a new HALT instruction will be added
 * to `program'. An HALT instruction doesn't require
 * any kind of parameter */
extern t_axe_instruction * gen_halt_instruction
      (t_program_infos *program);

/*----------------------------------------------------
 *                  LOAD/STORE OPERATIONS
 *---------------------------------------------------*/
 
/* A LOAD instruction requires the following parameters:
 * 1.  A destination register (where will be loaded the requested value)
 * 2.  A label information (can be a NULL pointer. If so, the addess
 *     value will be taken into consideration)
 * 3.  A direct address (if label is different from NULL) */
extern t_axe_instruction * gen_load_instruction
      (t_program_infos *program, int r_dest, t_axe_label *label, int address);

/* A STORE instruction copies a value from a register to a
 * specific memory location. The memory location can be
 * either a label identifier or a address reference.
 * In order to create a STORE instruction the caller must
 * privide a valid register location (`r_dest') and an
 * instance of `t_axe_label' or a numeric address */
extern t_axe_instruction * gen_store_instruction
      (t_program_infos *program, int r_dest, t_axe_label *label, int address);

extern t_axe_instruction * gen_store_array_instruction
      (t_program_infos *program,int r_source1,int immediate, t_axe_label *label, int address);

extern t_axe_instruction * gen_load_array_instruction
      (t_program_infos *program,int r_source1,int immediate, t_axe_label *label, int address);

extern t_axe_instruction * gen_jit_ldr_i_instruction
      (t_program_infos *program, int r_dest,int r_source1);

t_axe_instruction * gen_jit_sti_i_instruction
      (t_program_infos *program, int r_source1, int immediate, t_axe_label *label, int address);

/*----------------------------------------------------
 *                   READ/WRITE OPERATIONS
 *---------------------------------------------------*/
/* A READ instruction requires only one parameter:
 * A destination register (where will be loaded the value
 * read from standard input). */
extern t_axe_instruction * gen_read_instruction
               (t_program_infos *program, int r_dest);

/* A WRITE instruction requires only one parameter:
 * A destination register (where is located the value
 * that will be written to the standard output). */
extern t_axe_instruction * gen_write_instruction
               (t_program_infos *program, int r_dest);


/*----------------------------------------------------
 *                   BINARY OPERATIONS
 *---------------------------------------------------*/

/* ADDI Addition with immediate. */
extern t_axe_instruction * gen_addi_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate);

/* ADDCI Addition with immediate, generating carry. */
extern t_axe_instruction * gen_jit_addci_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate);

/* ADDXI Addition with immediate, using carry. Only after ADDCI. */
extern t_axe_instruction * gen_jit_addxi_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate);
/* ADDR Addition of registers. */
/* SUBI Subtraction with immediate. */
extern t_axe_instruction * gen_subi_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate);

/* SUBCI Subtraction with immediate, generating carry. */
extern t_axe_instruction * gen_jit_subci_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate);

/* SUBXI Subtraction with immediate, using carry. Only after SUBCI */
extern t_axe_instruction * gen_jit_subxi_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate);

/* RSBI Subtraction with immediate. */
extern t_axe_instruction * gen_jit_rsbi_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate);

/* ANDI Logical and with immediate. */
extern t_axe_instruction * gen_andli_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate);

/* ORI Logical or with immediate. */
extern t_axe_instruction * gen_orli_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate);

/* XORI Logical and with immediate. */
extern t_axe_instruction * gen_eorli_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate);

/* MULI Multiplication with immediate. */
extern t_axe_instruction * gen_muli_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate);

/* HMULI Multiplication with immediate. R1 <-- high bits of R2 * Immediate. */
extern t_axe_instruction * gen_jit_hmuli_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate);

/* DIVI Division with immediate. */
extern t_axe_instruction * gen_divi_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate);

/* MODI Modulo operation with immediate. */
extern t_axe_instruction * gen_jit_modi_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate);

/* LSHI Logical and with immediate. */
extern t_axe_instruction * gen_shli_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate);

/* RSHI Logical and with immediate. */
extern t_axe_instruction * gen_shri_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate);

/* LTI Set R1 if R2 < immediate. */
extern t_axe_instruction * gen_jit_lti_i_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate);

/* LEI Set R1 if R2 <= immediate. */
extern t_axe_instruction * gen_jit_lei_i_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate);

/* GTI Set R1 if R2 > immediate. */
extern t_axe_instruction * gen_jit_gti_i_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate);

/* GEI Set R1 if R2 >= immediate. */
extern t_axe_instruction * gen_jit_gei_i_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate);

/* EQI Set R1 if R2 = immediate. */
extern t_axe_instruction * gen_jit_eqi_i_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate);

/* NEI Set R1 if R2 != immediate. */
extern t_axe_instruction * gen_jit_nei_i_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate);

/* MOVI Move immediate*/
extern t_axe_instruction * gen_jit_movi_i_instruction
      (t_program_infos *program, int r_dest, int immediate);

/* NEGR*/
extern t_axe_instruction * gen_neg_instruction (t_program_infos *program
      , int r_dest, int r_source, int flags);

/*NOTR*/
t_axe_instruction * gen_notl_instruction
      (t_program_infos *program, int r_dest, int r_source1);

/*----------------------------------------------------
 *                   TERNARY OPERATIONS
 *---------------------------------------------------*/

/* ADDR Addition of registers. */
extern t_axe_instruction * gen_add_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags);

/* ADDCR Addition of registers generating carry. */
extern t_axe_instruction * gen_jit_addcr_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2);

/* ADDXR Addition of registers using carry. Only after ADDCR.*/
extern t_axe_instruction * gen_jit_addcr_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2);

/* SUBR Subtraction between registers. R1 <-- R2 - R3 */
extern t_axe_instruction * gen_sub_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags);

/* SUBCR Subtraction between registers generating carry. */
extern t_axe_instruction * gen_jit_subcr_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2);

/* SUBXR Subtraction between registers using carry. Only after SUBCR. */
extern t_axe_instruction * gen_jit_subxr_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2);

/* RSBR Subtraction between registers. R1 <-- R3 - R2 */
extern t_axe_instruction * gen_jit_rsbr_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2);

/* ANDR Logical and between registers. */
extern t_axe_instruction * gen_andl_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags);

/* ORR Logical or between registers. */
extern t_axe_instruction * gen_orl_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags);

/* XORR Logical xor between registers. */
extern t_axe_instruction * gen_eorl_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags);


/* MULR Multiplication between registers. */
extern t_axe_instruction * gen_mul_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags);

/* HMULR Multiplication between registers. R1 <-- high bits of R2 * R3. */
extern t_axe_instruction * gen_jit_hmulr_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2);

/* DIVR Division of two registers. */
extern t_axe_instruction * gen_div_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags);

/* MODR Modulo operation of registers. */
extern t_axe_instruction * gen_jit_modr_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2);

/* LSHR Left shift. R1 <-- R2 shifted to left by R3. */
extern t_axe_instruction * gen_shl_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags);

/* RSHR Right shift. R1 <-- R2 shifted to right by R3. */
extern t_axe_instruction * gen_shr_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags);

/* LTR Set R1 if R2 < R3 */
extern t_axe_instruction * gen_jit_ltr_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags);

/* LER Set R1 if R2 <= R3 */
extern t_axe_instruction * gen_jit_ler_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags);

/* GTR Set R1 if R2 > R3 */
extern t_axe_instruction * gen_jit_gtr_i_instruction (t_program_infos *program
      , int r_dest,int r_source1, int r_source2, int flags);

/* GER Set R1 if R2 >= R3 */
extern t_axe_instruction * gen_jit_ger_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags);

/* EQR Set R1 if R2 = R3 */
extern t_axe_instruction * gen_jit_eqr_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags);

/* NER Set R1 if R2 != R3 */
extern t_axe_instruction * gen_jit_ner_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags);


/*----------------------------------------------------
 *                   BINARY JUMP INSTRUCTIONS
 *---------------------------------------------------*/

/* BLTR Branch if R1 < R2.*/ 
extern t_axe_instruction * gen_jit_bltr_i_instruction
      (t_program_infos *program, int r_source1, int r_source2, t_axe_label *label, int addr);

/* BLER Branch if R1 <= R2.*/ 
extern t_axe_instruction * gen_jit_bler_i_instruction
      (t_program_infos *program,  int r_source1, int r_source2, t_axe_label *label, int addr);

/* BGTR Branch if R1 > R2.*/ 
extern t_axe_instruction * gen_jit_bgtr_i_instruction
      (t_program_infos *program,  int r_source1, int r_source2, t_axe_label *label, int addr);

/* BGER Branch if R1 <= R2.*/ 
extern t_axe_instruction * gen_jit_bger_i_instruction
      (t_program_infos *program,  int r_source1, int r_source2, t_axe_label *label, int addr);

/* BEQR Branch if R1 = R2.*/ 
extern t_axe_instruction * gen_jit_beqr_i_instruction
      (t_program_infos *program,  int r_source1, int r_source2, t_axe_label *label, int addr);

/* BNER Branch if R1 != R2.*/ 
extern t_axe_instruction * gen_jit_bner_i_instruction
      (t_program_infos *program,  int r_source1, int r_source2, t_axe_label *label, int addr);

/* BMSR Branch if R1 & R2 = 1.*/ 
extern t_axe_instruction * gen_jit_bmsr_i_instruction
      (t_program_infos *program,  int r_source1, int r_source2, t_axe_label *label, int addr);

/* BMcR Branch if !(R1 & R2) = 1.*/ 
extern t_axe_instruction * gen_jit_bmCr_i_instruction
      (t_program_infos *program,  int r_source1, int r_source2, t_axe_label *label, int addr);

/* BOADDR Branch if R1 += R2 overflows.*/ 
extern t_axe_instruction * gen_jit_boaddr_i_instruction
      (t_program_infos *program,  int r_source1, int r_source2, t_axe_label *label, int addr);

/* BOSUBR Branch if R1 -= R2 overflows.*/ 
extern t_axe_instruction * gen_jit_bsubr_i_instruction
      (t_program_infos *program,  int r_source1, int r_source2, t_axe_label *label, int addr);



/*----------------------------------------------------
 *           UNARY IMMEDIATE JUMP INSTRUCTIONS
 *---------------------------------------------------*/

/* BLTI Branch if R1 < immediate.*/ 
extern t_axe_instruction * gen_jit_blti_i_instruction
      (t_program_infos *program,  int r_source1, int immediate,t_axe_label *label, int addr);

/* BLEI Branch if R1 <= immediate.*/ 
extern t_axe_instruction * gen_jit_blei_i_instruction
      (t_program_infos *program,  int r_source1, int immediate,t_axe_label *label, int addr);

/* BGTI Branch if R1 > immediate.*/ 
extern t_axe_instruction * gen_jit_bgti_i_instruction
      (t_program_infos *program,  int r_source1, int immediate,t_axe_label *label, int addr);

/* BGEI Branch if R1 >= immediate.*/ 
extern t_axe_instruction * gen_jit_bgei_i_instruction
      (t_program_infos *program,  int r_source1, int immediate,t_axe_label *label, int addr);

/* BEQI Branch if R1 = immediate.*/ 
extern t_axe_instruction * gen_jit_beqi_i_instruction
      (t_program_infos *program,  int r_source1, int immediate,t_axe_label *label, int addr);

/* BNEI Branch if R1 != immediate.*/ 
extern t_axe_instruction * gen_jit_bnei_i_instruction
      (t_program_infos *program,  int r_source1, int immediate,t_axe_label *label, int addr);

/* BMSI Branch if R1 & immediate = 1.*/ 
extern t_axe_instruction * gen_jit_bmsi_i_instruction
      (t_program_infos *program,  int r_source1, int immediate,t_axe_label *label, int addr);

/* BMCI Branch if R1 & immediate = 0.*/ 
extern t_axe_instruction * gen_jit_bmci_i_instruction
      (t_program_infos *program,  int r_source1, int immediate,t_axe_label *label, int addr);

/* BOADDI Branch if R1 += immediate overflows.*/ 
extern t_axe_instruction * gen_jit_boaddi_i_instruction
      (t_program_infos *program,  int r_source1, int immediate,t_axe_label *label, int addr);

/* BOSUBI Branch if R1 -= immediate overflows.*/ 
extern t_axe_instruction * gen_jit_bosubi_i_instruction
      (t_program_infos *program,  int r_source1, int immediate,t_axe_label *label, int addr);



/*----------------------------------------------------
 *           UNCONDITIONAL JUMP INSTRUCTIONS
 *---------------------------------------------------*/

/* create a branch true instruction. By executing this instruction
 * the control is always passed to either the instruction with the label `label'
 * associated with, or (if `label' is a NULL pointer) to the explicit `address' */
extern t_axe_instruction * gen_bt_instruction
      (t_program_infos *program, t_axe_label *label, int addr);

/* create a branch true instruction. By executing this instruction
 * the control is always passed to the next instruction in the program
 * (i.e.: the instruction pointed by PC + 1). */
extern t_axe_instruction * gen_bf_instruction
      (t_program_infos *program, t_axe_label *label, int addr);


#endif
